package de.cmdp;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

public record EndComponent<S, A extends Choice<S>>(
        Set<S> states, Map<S, Collection<A>> choices, Map<S, Set<S>> successorCache) {
    public static <S, A extends Choice<S>> EndComponent<S, A> of(Set<S> states, Map<S, Collection<A>> choices) {
        assert states.equals(choices.keySet());
        // Every state has at least one action
        assert states.stream().map(choices::get).noneMatch(Collection::isEmpty);
        // Component is stable, all successors under all choices are subset of the state set
        assert states.stream()
                .map(choices::get)
                .flatMap(Collection::stream)
                .map(Choice::support)
                .allMatch(states::containsAll);

        return new EndComponent<>(states, choices, new HashMap<>());
    }

    public Collection<A> choices(S state) {
        assert states.contains(state);
        return choices.get(state);
    }

    public Set<S> successors(S state) {
        assert states.contains(state) : states;
        return successorCache.computeIfAbsent(state, s -> choices.get(s).stream()
                .map(Choice::support)
                .flatMap(Collection::stream)
                .collect(Collectors.toSet()));
    }

    public Iterator<S> successorsIterator(S state) {
        return successors(state).iterator();
    }

    @Override
    public boolean equals(Object obj) {
        return this == obj
                || (obj instanceof EndComponent<?, ?> component)
                        && states.equals(component.states)
                        && choices.equals(component.choices);
    }

    @Override
    public int hashCode() {
        return states.hashCode() + choices.hashCode();
    }

    @Override
    public String toString() {
        return "EC[%s]".formatted(states);
    }
}
