package de.cmdp;

import com.google.common.collect.Sets;
import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Deque;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.stream.Collectors;
import javax.annotation.Nullable;
import org.tinylog.Logger;

public class EndComponentAnalyser<S, A extends Choice<S>> {
    // TODO Tailored state search

    public List<EndComponent<S, A>> findComponents(Function<S, Collection<A>> model, Set<S> initial) {
        Logger.trace("Starting MECs search");

        List<EndComponent<S, A>> mecs = computeMaximalComponents(model, initial);
        if (mecs.isEmpty()) {
            Logger.trace("Found no MECs");
            return Collections.emptyList();
        }
        if (Logger.isTraceEnabled()) {
            Logger.trace(
                    "Found {} MECs:\n{}",
                    mecs.size(),
                    mecs.stream().map(Objects::toString).collect(Collectors.joining("\n")));
        }

        return mecs;
    }

    @Nullable
    private static <S, A extends Choice<S>> EndComponent<S, A> createMaximalSingleton(
            Function<S, Collection<A>> system, S state) {
        var choices = system.apply(state);
        if (choices.isEmpty()) {
            return null;
        }
        Set<S> stateSet = Set.of(state);
        var allowedActions =
                choices.stream().filter(c -> stateSet.equals(c.support())).toList();
        return allowedActions.isEmpty() ? null : EndComponent.of(stateSet, Map.of(state, allowedActions));
    }

    @Nullable
    private static <S, A extends Choice<S>> EndComponent<S, A> createMaximal(
            Function<S, Collection<A>> system, Set<S> states) {
        if (states.size() == 1) {
            S state = states.iterator().next();
            return createMaximalSingleton(system, state);
        }

        Map<S, Collection<A>> subActions = new HashMap<>(states.size());
        Set<S> stateSet = new HashSet<>(states);

        boolean changed;
        do {
            changed = false;

            Set<S> toRemove = new HashSet<>();
            for (S state : stateSet) {
                var stateRestriction = subActions.get(state);
                Predicate<A> allowedChoice = choice -> stateSet.containsAll(choice.support())
                        && Sets.intersection(toRemove, choice.support()).isEmpty();
                if (stateRestriction == null) {
                    var choices = system.apply(state);
                    if (choices.isEmpty()) {
                        toRemove.add(state);
                        changed = true;
                    } else {
                        List<A> actionCopy = new ArrayList<>();
                        boolean changedActions = false;
                        for (A choice : choices) {
                            if (allowedChoice.test(choice)) {
                                actionCopy.add(choice);
                            } else {
                                changedActions = true;
                            }
                        }
                        if (changedActions) {
                            if (actionCopy.isEmpty()) {
                                toRemove.add(state);
                            } else {
                                subActions.put(state, actionCopy);
                            }
                            changed = true;
                        } else {
                            subActions.put(state, actionCopy);
                        }
                    }
                } else {
                    if (stateRestriction.removeIf(allowedChoice.negate())) {
                        changed = true;
                        if (stateRestriction.isEmpty()) {
                            toRemove.add(state);
                            subActions.remove(state);
                        }
                    }
                }
            }
            stateSet.removeAll(toRemove);
            assert stateSet.equals(subActions.keySet());
        } while (changed);

        if (stateSet.isEmpty()) {
            return null;
        }

        return EndComponent.of(stateSet, subActions);
    }

    private static <S, A extends Choice<S>> List<EndComponent<S, A>> computeMaximalComponents(
            Function<S, Collection<A>> system, Set<S> initial) {
        List<EndComponent<S, A>> mecs = new ArrayList<>();

        Deque<EndComponent<S, A>> workList = new ArrayDeque<>();
        Function<S, Iterator<S>> successors = s -> system.apply(s).stream()
                .map(Choice::support)
                .flatMap(Collection::stream)
                .iterator();
        SccDecomposition.computeSccs(successors, initial, false, (Set<S> scc) -> {
            if (scc.size() == 1) {
                var component = createMaximalSingleton(system, scc.iterator().next());
                if (component == null) {
                    assert !SccDecomposition.isBscc(successors, scc);
                } else {
                    mecs.add(component);
                }
            } else {
                var component = createMaximal(system, scc);
                if (component != null) {
                    workList.add(component);
                }
            }
            return true;
        });

        List<EndComponent<S, A>> preMecs = new ArrayList<>();

        while (!workList.isEmpty()) {
            var mec = workList.remove();
            preMecs.clear();

            SccDecomposition.computeSccs(mec::successorsIterator, mec.states(), false, scc -> {
                var component = createMaximal(system, scc);
                if (component != null) {
                    preMecs.add(component);
                }
                return true;
            });

            if (preMecs.size() == 1) {
                var refinedEc = preMecs.get(0);
                assert !refinedEc.states().isEmpty();
                if (mec.equals(refinedEc)) {
                    mecs.add(refinedEc);
                    continue;
                }
            }
            workList.addAll(preMecs);
        }
        assert mecs.size() == Set.copyOf(mecs).size();

        return mecs;
    }
}
