package de.cmdp;

import static com.google.common.base.Preconditions.checkState;

import com.google.common.base.Function;
import com.google.common.collect.Lists;
import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.Deque;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;
import java.util.function.BiFunction;
import java.util.function.Predicate;
import org.tinylog.Logger;

public class GlobalExplorer {
    private final Map<GlobalState, List<GlobalAction>> globalGraph = new HashMap<>();
    private final Map<GlobalState, Deque<PartialAction>> globalActions = new HashMap<>();
    private final Set<GlobalState> globalExplored = new HashSet<>();
    private final Queue<GlobalState> globalQueue = new ArrayDeque<>();
    private final Set<GlobalState> goalStates = new HashSet<>();
    private final List<ModuleGraph> moduleGraphs;
    private final Function<String, Set<ModuleGraph>> labelModules;
    private final BiFunction<GlobalState, String, Object> valuation;
    private final Predicate<GlobalState> goal;
    private final Map<GlobalState, Set<String>> availableLabels = new HashMap<>();
    private final Comparator<String> labelPriority;

    public GlobalExplorer(
            List<ModuleGraph> moduleGraphs,
            Function<String, Set<ModuleGraph>> labelModules,
            BiFunction<GlobalState, String, Object> valuation,
            Predicate<GlobalState> goal,
            Comparator<String> labelPriority) {
        this.moduleGraphs = moduleGraphs;
        this.labelModules = labelModules;
        this.valuation = valuation;
        this.goal = goal;
        this.labelPriority = labelPriority;
    }

    public boolean explore(GlobalState initialState) {
        globalExplored.add(initialState);
        globalQueue.add(initialState);
        return explore();
    }

    private boolean explore() {
        boolean changed = false;
        while (!globalQueue.isEmpty()) {
            var state = globalQueue.poll();
            globalGraph.computeIfAbsent(state, k -> new ArrayList<>());

            if (goal.test(state)) {
                goalStates.add(state);
                continue;
            }

            if (exploreNext(state)) {
                changed = true;
            }
        }
        assert globalGraph.keySet().equals(globalExplored);

        assert globalGraph.values().stream()
                .flatMap(List::stream)
                .map(GlobalAction::distribution)
                .map(Distribution::map)
                .map(Map::keySet)
                .flatMap(Collection::stream)
                .allMatch(globalGraph::containsKey);
        return changed;
    }

    private Deque<PartialAction> computeActions(GlobalState state) {
        Map<ModuleGraph, List<ModuleGraph.ModuleAction>> availableActions = new HashMap<>();
        Map<String, Set<ModuleGraph>> enabledByLabel = new HashMap<>();

        for (ModuleGraph module : moduleGraphs) {
            var moduleTransitions = module.transitions.get(state.moduleState(module));

            List<ModuleGraph.ModuleAction> actions = new ArrayList<>();
            Predicate<Map.Entry<String, Object>> readEnabled =
                    read -> valuation.apply(state, read.getKey()).equals(read.getValue());

            // state.moduleState(variableToModule.get(read.getKey())).value(read.getKey()).equals(read.getValue());

            for (var transition : moduleTransitions) {
                if (transition.foreignReads().entrySet().stream().allMatch(readEnabled)) {
                    boolean added = enabledByLabel
                            .computeIfAbsent(transition.label(), k -> new HashSet<>())
                            .add(module);
                    checkState(
                            added,
                            "Multiple actions for label %s from module %s in state %s",
                            transition.label(),
                            module,
                            state);
                    actions.add(transition);
                }
            }
            availableActions.put(module, actions);
        }

        Set<Object> enabledLabels = new HashSet<>();
        for (var entry : enabledByLabel.entrySet()) {
            if (labelModules.apply(entry.getKey()).equals(entry.getValue())) {
                enabledLabels.add(entry.getKey());
            }
        }
        for (var moduleActions : availableActions.values()) {
            moduleActions.removeIf(a -> !enabledLabels.contains(a.label()));
        }
        availableActions.values().removeIf(Collection::isEmpty);

        Map<String, Map<ModuleGraph, ModuleGraph.ModuleAction>> labelledActions = new HashMap<>();
        for (var entry : availableActions.entrySet()) {
            for (ModuleGraph.ModuleAction moduleAction : entry.getValue()) {
                var previous = labelledActions
                        .computeIfAbsent(moduleAction.label(), k -> new HashMap<>())
                        .put(entry.getKey(), moduleAction);
                assert previous == null;
            }
        }

        availableLabels.put(state, Set.copyOf(labelledActions.keySet()));

        List<PartialAction> actions = new ArrayList<>();
        for (ModuleGraph moduleGraph : moduleGraphs) {
            for (ModuleGraph.ModuleAction action : availableActions.getOrDefault(moduleGraph, List.of())) {
                var moduleActions = labelledActions.remove(action.label());
                if (moduleActions != null) {
                    actions.add(new PartialAction(action.label(), moduleActions));
                }
            }
        }
        actions.sort(Comparator.comparing(a -> a.label, labelPriority));
        return new ArrayDeque<>(actions);
    }

    public Set<String> availableLabels(GlobalState state) {
        return availableLabels.get(state);
    }

    public boolean exploreNext(GlobalState state) {
        assert globalExplored.contains(state);
        Deque<PartialAction> actions = globalActions.computeIfAbsent(state, this::computeActions);
        if (actions.isEmpty()) {
            Logger.trace("No available actions in state {}", state);
            return false;
        }

        Map<GlobalState, Double> distribution = new HashMap<>();
        Map<ModuleGraph, ModuleGraph.ModuleState> globalMap = new HashMap<>(state.states());
        PartialAction nextAction = actions.pop();
        String label = nextAction.label();

        List<ModuleGraph> modules = new ArrayList<>();
        List<List<Map.Entry<ModuleGraph.ModuleState, Double>>> effects = new ArrayList<>();
        for (var entry : nextAction.actions().entrySet()) {
            assert label.equals(entry.getValue().label());
            modules.add(entry.getKey());
            effects.add(List.copyOf(entry.getValue().distribution().map().entrySet()));
        }

        for (var effect : Lists.cartesianProduct(effects)) {
            double probability = 1.0;
            for (int i = 0; i < effect.size(); i++) {
                var entry = effect.get(i);
                globalMap.put(modules.get(i), entry.getKey());
                probability *= entry.getValue();
            }
            distribution.put(new GlobalState(globalMap), probability);
        }
        GlobalAction newAction = new GlobalAction(label, Distribution.of(distribution));

        Logger.trace("Progressing with {} in state {}", newAction, state);

        for (GlobalState successor : newAction.distribution().support()) {
            if (globalExplored.add(successor)) {
                globalQueue.add(successor);
            }
        }
        globalGraph.get(state).add(newAction);
        return true;
    }

    public Set<GlobalState> states() {
        return globalExplored;
    }

    public List<GlobalAction> actions(GlobalState state) {
        assert globalExplored.contains(state);
        return globalGraph.getOrDefault(state, List.of());
    }

    public Set<GlobalState> goal() {
        return goalStates;
    }

    public void drop(GlobalState state, Predicate<PartialAction> filter) {
        globalActions.get(state).removeIf(filter);
    }

    public record PartialAction(String label, Map<ModuleGraph, ModuleGraph.ModuleAction> actions) {}
}
