package de.cmdp;

import static com.google.common.base.Preconditions.checkState;

import com.google.common.collect.Sets;
import de.cmdp.ModuleGraph.ModuleState;
import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.Deque;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.concurrent.Callable;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.stream.Collectors;
import org.ojalgo.optimisation.ExpressionsBasedModel;
import org.ojalgo.optimisation.Optimisation;
import org.ojalgo.optimisation.Variable;
import org.tinylog.Logger;
import parser.EvaluateContext;
import parser.PrismParser;
import parser.Values;
import parser.ast.Declaration;
import parser.ast.DeclarationBool;
import parser.ast.DeclarationInt;
import parser.ast.Expression;
import parser.ast.ModulesFile;
import parser.type.TypeBool;
import picocli.CommandLine;
import prism.PrismException;
import prism.PrismLangException;
import prism.UndefinedConstants;

@CommandLine.Command(name = "cd", mixinStandardHelpOptions = true)
public class Main implements Callable<Integer> {
    @CommandLine.Option(
            names = {"--model", "-m"},
            required = true)
    private Path modelPath;

    @CommandLine.Option(
            names = {"--property"},
            required = true)
    private String property;

    @CommandLine.Option(
            names = {"--const", "-c"},
            split = ",")
    private List<String> constants = new ArrayList<>();

    @CommandLine.Option(names = "--assume-cd")
    private boolean assumeCausallyDeterministic = false;

    @CommandLine.Option(names = "--render")
    private boolean render = false;

    public static void main(String[] args) {
        System.exit(new CommandLine(new Main()).execute(args));
    }

    @Override
    public Integer call() throws PrismException, IOException {
        // Prism prism = new Prism(new PrismDevNullLog());
        PrismParser parser = new PrismParser();

        ModulesFile modulesFile;
        try (InputStream in = new BufferedInputStream(Files.newInputStream(modelPath))) {
            modulesFile = parser.parseModulesFile(in, null);
        }
        modulesFile.tidyUp();
        UndefinedConstants undefinedConstants = new UndefinedConstants(modulesFile, null);
        if (constants.isEmpty()) {
            undefinedConstants.checkAllDefined();
        } else {
            undefinedConstants.defineUsingConstSwitch(String.join(",", constants));
        }
        modulesFile.setSomeUndefinedConstants(undefinedConstants.getMFConstantValues());
        undefinedConstants.iterateModel();

        Expression expression = parser.parseSingleExpression(new ByteArrayInputStream(property.getBytes()));
        expression = (Expression) expression.findAllVars(modulesFile.getAllVars(), modulesFile.getVarTypes());
        expression = (Expression) expression.findAllConstants(modulesFile.getConstantList());
        expression = (Expression) expression.findAllFormulas(modulesFile.getFormulaList());
        expression = (Expression) expression.expandFormulas(modulesFile.getFormulaList());
        expression = (Expression) expression.expandLabels(modulesFile.getLabelList());
        expression = (Expression) expression.evaluatePartially(modulesFile.getConstantValues());
        expression = (Expression) expression.expandConstants(modulesFile.getConstantList());
        expression.checkValid(modulesFile.getModelType());
        expression.typeCheck();
        expression.semanticCheck();
        if (!(expression.getType() instanceof TypeBool)) {
            System.err.printf(
                    "Expression %s is not a positional formula, but of type %s%n", expression, expression.getType());
            return 1;
        }

        // TODO Globals are just another module
        if (modulesFile.getNumGlobals() > 0) {
            System.err.println("Global variables");
            return 1;
        }

        // TODO Approximate: All states for which the formula is not false
        if (modulesFile.getInitialStates() != null) {
            System.err.println("Multiple initial states");
            return 1;
        }

        Values constantValues = modulesFile.getConstantValues();
        int moduleCount = modulesFile.getNumModules();
        Map<String, String> variableToModuleName = new HashMap<>();
        Map<String, List<Object>> variableRanges = new HashMap<>();

        for (int i = 0; i < moduleCount; i++) {
            var module = modulesFile.getModule(i);
            String moduleName = module.getName();
            for (Declaration declaration : module.getDeclarations()) {
                String variableName = declaration.getName();
                checkState(!variableToModuleName.containsKey(variableName));
                variableToModuleName.put(variableName, moduleName);
                var type = declaration.getDeclType();

                List<Object> range = new ArrayList<>();
                if (type instanceof DeclarationInt integer) {
                    int low = integer.getLow().evaluateInt(constantValues);
                    int high = integer.getHigh().evaluateInt(constantValues);
                    for (int v = low; v <= high; v++) {
                        range.add(v);
                    }
                } else if (type instanceof DeclarationBool) {
                    range.add(Boolean.FALSE);
                    range.add(Boolean.TRUE);
                } else {
                    throw new IllegalArgumentException(
                            "Unsupported declaration type " + type.getClass().getSimpleName());
                }
                variableRanges.put(variableName, List.copyOf(range));
            }
        }

        List<ModuleGraph> moduleGraphs = new ArrayList<>();
        for (int i = 0; i < moduleCount; i++) {
            var module = modulesFile.getModule(i);
            moduleGraphs.add(ModuleGraph.build(i, module, constantValues, variableRanges));
        }
        Map<String, ModuleGraph> variableToModule = new HashMap<>();
        for (ModuleGraph module : moduleGraphs) {
            for (String variable : module.variables) {
                variableToModule.put(variable, module);
            }
        }

        if (!assumeCausallyDeterministic) {
            boolean error = false;
            for (ModuleGraph module : moduleGraphs) {
                var moduleName = module.module.getName();

                for (var entry : module.transitions.entrySet()) {
                    var state = entry.getKey();

                    for (var actionA : entry.getValue()) {
                        if (actionA.foreignReads().isEmpty()
                                && actionA.distribution().isOnlySuccessor(state)) {
                            continue;
                        }
                        for (var actionB : entry.getValue()) {
                            if (actionA == actionB) {
                                continue;
                            }
                            if (actionB.foreignReads().isEmpty()
                                    && actionB.distribution().isOnlySuccessor(state)) {
                                continue;
                            }

                            Set<String> variables =
                                    new HashSet<>(actionA.foreignReads().keySet());
                            variables.retainAll(actionB.foreignReads().keySet());

                            boolean equal = true;
                            for (String variable : variables) {
                                if (!actionA.foreignReads()
                                        .get(variable)
                                        .equals(actionB.foreignReads().get(variable))) {
                                    equal = false;
                                    break;
                                }
                            }
                            if (equal) {
                                Map<String, Object> values = new HashMap<>();
                                for (String variable : variables) {
                                    values.put(variable, actionA.foreignReads().get(variable));
                                }
                                System.err.printf(
                                        "Module %s, state %s, can take actions %s(%s) and %s(%s) in context %s%n",
                                        moduleName, state, actionA.label(), actionA, actionB.label(), actionB, values);
                                error = true;
                            }
                        }
                    }
                    Set<Set<ModuleGraph>> stateParticipating = new HashSet<>();
                    for (ModuleGraph.ModuleAction action : entry.getValue()) {
                        Set<ModuleGraph> involvedModules = new HashSet<>();
                        involvedModules.add(module);
                        for (String variable : action.foreignReads().keySet()) {
                            involvedModules.add(variableToModule.get(variable));
                        }
                        stateParticipating.add(Set.copyOf(involvedModules));
                    }

                    if (stateParticipating.size() > 1) {
                        Logger.error(
                                "CD violation: Module {}, state {} has loc sets: {}",
                                moduleName,
                                state,
                                stateParticipating);
                        error = true;
                    } else {
                        Logger.trace("Module {}, state {} has loc sets: {}", moduleName, state, stateParticipating);
                    }
                }
            }

            if (error) {
                System.err.println("Model is not causally deterministic");
                return 1;
            }
        }
        Map<String, Set<ModuleGraph>> labelModules = new HashMap<>();
        for (ModuleGraph module : moduleGraphs) {
            for (var actions : module.transitions.values()) {
                for (ModuleGraph.ModuleAction action : actions) {
                    labelModules
                            .computeIfAbsent(action.label(), k -> new HashSet<>())
                            .add(module);
                }
            }
        }

        Map<ModuleGraph, Map<ModuleState, Set<ModuleGraph>>> dependencies = new HashMap<>();
        for (ModuleGraph module : moduleGraphs) {
            Map<ModuleState, Set<ModuleGraph>> stateDependencies = new HashMap<>();
            Set<ModuleGraph> involvedModules = new HashSet<>();
            for (var entry : module.transitions.entrySet()) {
                var state = entry.getKey();
                involvedModules.add(module);
                for (ModuleGraph.ModuleAction action : entry.getValue()) {
                    for (String variable : action.foreignReads().keySet()) {
                        involvedModules.add(variableToModule.get(variable));
                    }
                    involvedModules.addAll(labelModules.get(action.label()));
                }
                stateDependencies.put(state, Set.copyOf(involvedModules));
            }
            dependencies.put(module, Map.copyOf(stateDependencies));
        }

        List<ModuleGraph> focusModules = new ArrayList<>();
        Set<String> focusVariables = Set.copyOf(expression.getAllVars());
        Logger.trace("Expression {} has variables {}", expression, focusVariables);
        for (ModuleGraph moduleGraph : moduleGraphs) {
            if (Sets.intersection(moduleGraph.variables, focusVariables).isEmpty()) {
                continue;
            }
            focusModules.add(moduleGraph);
        }
        if (focusModules.isEmpty()) {
            System.err.printf("Trivial property %s%n", expression);
            return 1;
        }
        if (focusModules.size() > 1) {
            Logger.error(
                    "CD violation: Expression {} relates to multiple modules {}",
                    expression,
                    focusModules.stream().map(Objects::toString).collect(Collectors.joining(", ")));
            System.err.println("Query is not casually deterministic");
            return 1;
        }
        ModuleGraph focusModule = focusModules.get(0);

        List<ModuleGraph> modulePriority = new ArrayList<>();
        Set<ModuleGraph> priorityExploration = new HashSet<>(Set.of(focusModule));
        Deque<ModuleGraph> priorityExplorationQueue = new ArrayDeque<>(priorityExploration);
        while (!priorityExplorationQueue.isEmpty()) {
            var module = priorityExplorationQueue.pollLast();
            modulePriority.add(module);
            for (Set<ModuleGraph> stateDependencies : dependencies.get(module).values()) {
                for (ModuleGraph successor : stateDependencies.stream()
                        .sorted(Comparator.<ModuleGraph>comparingInt(m -> m.labels.size())
                                .thenComparing(m -> m.module.getName()))
                        .toList()) {
                    if (priorityExploration.add(successor)) {
                        priorityExplorationQueue.add(successor);
                    }
                }
            }
        }
        Map<ModuleGraph, Integer> modulePriorityMap = new HashMap<>();
        // modulePriority = List.copyOf(Lists.reverse(modulePriority));
        modulePriority.forEach(m -> modulePriorityMap.put(m, modulePriorityMap.size()));

        Map<String, Integer> labelPriority = new HashMap<>();
        for (ModuleGraph moduleGraph : modulePriority) {
            List<String> moduleLabels = new ArrayList<>(moduleGraph.labels);
            moduleLabels.sort(Comparator.<String>comparingInt(l -> labelModules.get(l).stream()
                            .map(modulePriorityMap::get)
                            .mapToInt(Integer::intValue)
                            .max()
                            .orElseThrow())
                    .thenComparing(Function.identity()));
            moduleLabels.forEach(l -> labelPriority.putIfAbsent(l, labelPriority.size()));
        }

        Logger.debug("Module priority: {}", modulePriority);
        Logger.debug(
                "Label priority: {}",
                labelPriority.entrySet().stream()
                        .sorted(Map.Entry.comparingByValue())
                        .map(Map.Entry::getKey)
                        .collect(Collectors.joining(",")));

        GlobalState initialState = new GlobalState(
                moduleGraphs.stream().collect(Collectors.toMap(Function.identity(), ModuleGraph::initialState)));
        Expression goalExpression = expression;
        Predicate<GlobalState> goalPredicate = s -> {
            try {
                return goalExpression.evaluateBoolean(new Evaluate(s, variableToModule));
            } catch (PrismLangException e) {
                throw new RuntimeException(e);
            }
        };
        GlobalExplorer explorer = new GlobalExplorer(
                moduleGraphs,
                labelModules::get,
                (s, v) -> s.moduleState(variableToModule.get(v)).value(v),
                goalPredicate,
                Comparator.comparing(labelPriority::get));
        explorer.explore(initialState);
        EndComponentAnalyser<GlobalState, GlobalAction> analyser = new EndComponentAnalyser<>();
        Set<GlobalState> modifiedStates = new HashSet<>(Set.of(initialState));
        while (true) {
            var components = analyser.findComponents(explorer::actions, modifiedStates);
            modifiedStates.clear();

            boolean any = false;
            for (EndComponent<GlobalState, GlobalAction> component : components) {
                if (component.states().stream().anyMatch(goalPredicate)) {
                    continue;
                }
                if (!component.states().stream()
                        .map(explorer::actions)
                        .flatMap(Collection::stream)
                        .map(GlobalAction::support)
                        .allMatch(component.states()::containsAll)) {
                    // Not a bottom component
                    continue;
                }

                Map<String, List<GlobalState>> available = new HashMap<>();
                Set<String> used = new HashSet<>();
                for (GlobalState state : component.states()) {
                    for (String label : explorer.availableLabels(state)) {
                        available.computeIfAbsent(label, k -> new ArrayList<>()).add(state);
                    }
                    for (GlobalAction action : explorer.actions(state)) {
                        used.add(action.label());
                    }
                }
                available.keySet().removeAll(used);
                if (available.isEmpty()) {
                    continue;
                }
                any = true;
                for (GlobalState state : component.states()) {
                    explorer.drop(state, a -> !available.containsKey(a.label()));
                }
                var bestLabel = available.keySet().stream()
                        .min(Comparator.comparing(labelPriority::get))
                        .orElseThrow();
                // available.keySet().iterator().next();

                var state = available.get(bestLabel).iterator().next();
                Logger.trace("Progressing delayed action {}", bestLabel);
                boolean changed = explorer.explore(state);
                modifiedStates.addAll(component.states());
                assert changed;
            }
            if (!any) {
                break;
            }
        }

        var goalStates = explorer.goal();
        Map<GlobalState, Integer> numbers = new HashMap<>();
        explorer.states().forEach(s -> numbers.put(s, numbers.size()));

        /* globalGraph.forEach((s, ss) -> {
            System.out.println(numbers.get(s) + (goalStates.contains(s) ? "G" : "") + "   " + s);
            ss.forEach(a -> {
                System.out.println(a.label + " " + a.distribution.map().entrySet().stream()
                        .collect(Collectors.toMap(e -> numbers.get(e.getKey()), Map.Entry::getValue)));
            });
        }); */

        Logger.info("Model has {} states", explorer.states().size());
        if (render) {
            Render.display(explorer, Set.of(initialState), state -> {
                if (goalStates.contains(state)) {
                    return "goal";
                }
                List<String> labels = new ArrayList<>();
                for (ModuleGraph graph : moduleGraphs) {
                    state.moduleState(graph).values().entrySet().stream()
                            .sorted(Map.Entry.comparingByKey())
                            .forEach(e -> labels.add(e.getValue().toString()));
                }
                return String.join(",", labels);
            });
        }

        Optimisation.Options options = new Optimisation.Options();
        options.sparse = true;
        // options.solution = NumberContext.of(12);
        options.validate = false;
        ExpressionsBasedModel model = new ExpressionsBasedModel(options);
        Map<GlobalState, Variable> stateVariables = new HashMap<>();
        explorer.states()
                .forEach(s -> stateVariables.put(
                        s, model.addVariable().weight(1).lower(0).upper(1)));
        // stateVariables.values().forEach(v -> v.setValue(1.0));
        for (GlobalState goalState : goalStates) {
            stateVariables.get(goalState).level(1.0).setValue(1.0);
        }
        explorer.states().forEach(s -> {
            var stateVar = stateVariables.get(s);
            explorer.actions(s).forEach(action -> {
                var actionExpression = model.addExpression().set(stateVar, -1).upper(0);
                action.distribution().map().forEach((t, p) -> actionExpression.add(stateVariables.get(t), p));
            });
        });
        Optimisation.Result minimise = model.minimise();
        Map<GlobalState, Double> stateValues = new HashMap<>();
        for (var entry : stateVariables.entrySet()) {
            stateValues.put(entry.getKey(), minimise.doubleValue(model.indexOf(entry.getValue())));
        }

        System.out.printf("Value in initial state: %.16f%n", stateValues.get(initialState));
        return 0;
    }

    private static class Evaluate extends EvaluateContext {
        private final GlobalState state;
        private final Map<String, ModuleGraph> variableToModule;

        public Evaluate(GlobalState state, Map<String, ModuleGraph> variableToModule) {
            this.state = state;
            this.variableToModule = variableToModule;
        }

        @Override
        public Object getVarValue(String s, int i) {
            return state.moduleState(variableToModule.get(s)).value(s);
        }
    }
}
