package de.cmdp;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkState;

import com.google.common.base.Strings;
import com.google.common.collect.Lists;
import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;
import java.util.stream.Collectors;
import parser.EvaluateContext;
import parser.Values;
import parser.ast.ASTElement;
import parser.ast.Command;
import parser.ast.Declaration;
import parser.ast.Expression;
import parser.ast.ExpressionLiteral;
import parser.ast.Module;
import parser.ast.Update;
import parser.ast.UpdateElement;
import prism.PrismLangException;

public class ModuleGraph {
    public ModuleState initialState() {
        return initialState;
    }

    public record ModuleState(Map<String, Object> values) {
        public Object value(String name) {
            return values.get(name);
        }

        @Override
        public String toString() {
            return "["
                    + values.entrySet().stream()
                            .sorted(Map.Entry.comparingByKey())
                            .map(e -> e.getKey() + ":" + e.getValue())
                            .collect(Collectors.joining(","))
                    + "]";
        }
    }

    public record ModuleAction(String label, Map<String, Object> foreignReads, Distribution<ModuleState> distribution) {
        @Override
        public String toString() {
            return label;
        }
    }

    public final Module module;
    public final Set<ModuleState> states;
    public final Map<ModuleState, Set<ModuleAction>> transitions;
    public final Set<String> variables;
    public final ModuleState initialState;
    public final int index;
    public final Set<String> labels;

    private ModuleGraph(
            Module module,
            Set<ModuleState> states,
            Map<ModuleState, Set<ModuleAction>> transitions,
            Set<String> variables,
            ModuleState initialState,
            int index) {
        this.module = module;
        this.states = Set.copyOf(states);
        this.transitions = Map.copyOf(transitions);
        this.variables = Set.copyOf(variables);
        this.initialState = initialState;
        this.index = index;

        Set<String> labels = new HashSet<>();
        transitions.values().forEach(as -> as.forEach(a -> labels.add(a.label)));
        this.labels = Set.copyOf(labels);

        checkArgument(states.containsAll(transitions.keySet()));
        for (ModuleState state : states) {
            checkArgument(state.values.keySet().equals(variables));
        }
        for (Set<ModuleAction> actions : transitions.values()) {
            for (ModuleAction action : actions) {
                checkArgument(states.containsAll(action.distribution.support()));
            }
        }
    }

    public static ModuleGraph build(
            int index, Module module, Values constants, Map<String, List<Object>> variableRanges)
            throws PrismLangException {
        // TODO Figure this out
        checkArgument(module.getInvariant() == null);

        Set<String> moduleVariables = new HashSet<>();
        Map<ModuleState, Set<ModuleAction>> transitions = new HashMap<>();

        Map<String, Object> initialStateValues = new HashMap<>();
        for (Declaration declaration : module.getDeclarations()) {
            moduleVariables.add(declaration.getName());
            initialStateValues.put(
                    declaration.getName(), declaration.getStartOrDefault().evaluate(constants));
        }

        ModuleState initialState = new ModuleState(Map.copyOf(initialStateValues));
        Set<ModuleState> states = new HashSet<>(List.of(initialState));
        Queue<ModuleState> queue = new ArrayDeque<>(states);

        while (!queue.isEmpty()) {
            var state = queue.poll();
            var stateValues = new StateEvaluateContext(state, constants);
            Set<ModuleAction> stateActions = new HashSet<>();

            for (Command command : module.getCommands()) {
                ASTElement simplify = command.getGuard()
                        .deepCopy()
                        .evaluatePartially(stateValues)
                        .simplify();
                if (ExpressionLiteral.False().equals(simplify)) {
                    continue;
                }
                List<String> foreignVariables = List.copyOf(simplify.getAllVars());

                var updates = command.getUpdates();
                List<PartialUpdate> partialUpdates = new ArrayList<>();
                for (int i = 0; i < updates.getUpdates().size(); i++) {
                    Update update = updates.getUpdate(i);
                    Expression probability = updates.getProbability(i);
                    partialUpdates.add(new PartialUpdate(
                            (Update) update.deepCopy()
                                    .evaluatePartially(stateValues)
                                    .simplify(),
                            probability == null
                                    ? ExpressionLiteral.Double(1.0)
                                    : (Expression) probability.deepCopy().evaluatePartially(stateValues)));
                }

                int count = foreignVariables.size();
                List<List<Object>> valueRanges = new ArrayList<>(count);
                Map<String, Integer> variableIndexMap = new HashMap<>(count);

                for (String variable : foreignVariables) {
                    valueRanges.add(variableRanges.get(variable));
                    variableIndexMap.put(variable, variableIndexMap.size());
                }

                for (List<Object> foreignAssignment : Lists.cartesianProduct(valueRanges)) {
                    EvaluateContext context = new ExternalEvaluation(variableIndexMap, foreignAssignment);
                    ASTElement evaluated =
                            simplify.deepCopy().evaluatePartially(context).simplify();
                    if (ExpressionLiteral.False().equals(evaluated)) {
                        continue;
                    }
                    checkState(ExpressionLiteral.True().equals(evaluated), evaluated);

                    Map<ModuleState, Double> distribution = new HashMap<>();
                    for (PartialUpdate update : partialUpdates) {
                        Map<String, Object> successor = new HashMap<>(state.values);
                        for (UpdateElement element : ((Update) update.update.evaluatePartially(context))) {
                            successor.put(
                                    element.getVar(), element.getExpression().evaluate());
                        }
                        double probability = (Double) ((ExpressionLiteral) update.probability
                                        .evaluatePartially(context)
                                        .simplify())
                                .getValue();
                        distribution.put(new ModuleState(successor), probability);
                    }
                    Map<String, Object> foreignReads = new HashMap<>();
                    for (String variable : foreignVariables) {
                        if (!moduleVariables.contains(variable)) {
                            foreignReads.put(variable, foreignAssignment.get(variableIndexMap.get(variable)));
                        }
                    }

                    String label = Strings.isNullOrEmpty(command.getSynch())
                            ? "t_%s_%d_%d".formatted(module.getName(), command.getBeginLine(), command.getBeginColumn())
                            : command.getSynch();
                    stateActions.add(new ModuleAction(label, foreignReads, Distribution.of(distribution)));
                }
            }

            transitions.put(state, Set.copyOf(stateActions));
            for (ModuleAction action : stateActions) {
                for (ModuleState moduleState : action.distribution.support()) {
                    if (states.add(moduleState)) {
                        queue.add(moduleState);
                    }
                }
            }
        }

        return new ModuleGraph(module, states, transitions, moduleVariables, initialState, index);
    }

    private record PartialUpdate(Update update, Expression probability) {}

    private static final class StateEvaluateContext extends EvaluateContext {
        private final ModuleState state;

        public StateEvaluateContext(ModuleState state, Values constants) {
            this.state = state;
            this.setConstantValues(constants);
        }

        @Override
        public Object getVarValue(String s, int i) {
            return state.values.get(s);
        }
    }

    private static final class ExternalEvaluation extends EvaluateContext {
        private final Map<String, Integer> index;
        private final List<Object> assignment;

        public ExternalEvaluation(Map<String, Integer> index, List<Object> assignment) {
            this.index = index;
            this.assignment = assignment;
        }

        @Override
        public Object getVarValue(String s, int i) {
            return assignment.get(index.get(s));
        }
    }

    @Override
    public boolean equals(Object obj) {
        return this == obj
                || obj instanceof ModuleGraph graph && module.getName().equals(graph.module.getName());
    }

    @Override
    public int hashCode() {
        return module.getName().hashCode();
    }

    @Override
    public String toString() {
        return module.getName();
    }
}
