package de.cmdp;

import com.google.common.collect.Iterables;
import com.google.common.collect.Iterators;
import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Deque;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.function.Function;
import java.util.function.Predicate;

/**
 * Finds the SCCs of a given graph / transition system using Tarjan's algorithm. Adapted from owl.
 */
public final class SccDecomposition<S> {
    // TODO Parallel?
    // TODO Optimize for only BSSCs

    // Initial value for the low link - since we update the low-link whenever we find a link to a
    // state we can use this to detect trivial SCCs. MAX_VALUE is important for "<" comparisons
    static final int NO_LINK = Integer.MAX_VALUE;

    private final Deque<S> explorationStack = new ArrayDeque<>();
    private final boolean includeTransient;
    private final Predicate<Set<S>> consumer;
    private final Deque<TarjanState<S>> path = new ArrayDeque<>();
    private final Set<S> processedNodes = new HashSet<>();
    private final Map<S, TarjanState<S>> stateMap = new HashMap<>();
    private final Function<S, ? extends Iterator<S>> successorFunction;
    private int index = 0;

    private SccDecomposition(
            Function<S, ? extends Iterator<S>> successorFunction,
            boolean includeTransient,
            Predicate<Set<S>> consumer) {
        this.successorFunction = successorFunction;
        this.includeTransient = includeTransient;
        this.consumer = consumer;
    }

    public static <S> void computeSccs(
            Function<S, ? extends Iterator<S>> function,
            Set<S> initialStates,
            boolean includeTransient,
            Predicate<Set<S>> consumer) {
        SccDecomposition<S> decomposition = new SccDecomposition<>(function, includeTransient, consumer);
        initialStates.forEach((S initialState) -> {
            if (!decomposition.stateMap.containsKey(initialState)
                    && !decomposition.processedNodes.contains(initialState)) {
                decomposition.run(initialState);
            }
        });
    }

    public static <S> List<Set<S>> computeSccs(
            Function<S, ? extends Iterator<S>> function, Set<S> initialStates, boolean includeTransient) {
        List<Set<S>> sccs = new ArrayList<>();
        computeSccs(function, initialStates, includeTransient, sccs::add);
        assert includeTransient || sccs.stream().noneMatch(scc -> isTransient(function, scc));
        return sccs;
    }

    public static <S> boolean isTransient(Function<S, ? extends Iterator<S>> function, Set<S> scc) {
        if (scc.size() > 1) {
            return false;
        }

        var state = Iterables.getOnlyElement(scc);
        return !Iterators.contains(function.apply(state), state);
    }

    public static <S, A extends Choice<S>> boolean isBscc(Function<S, ? extends Iterator<S>> function, Set<S> scc) {
        return scc.stream().parallel().noneMatch(state -> {
            Iterator<S> iterator = function.apply(state);
            while (iterator.hasNext()) {
                if (!scc.contains(iterator.next())) {
                    return true;
                }
            }
            return false;
        });
    }

    private TarjanState<S> create(S node) {
        assert !stateMap.containsKey(node) && !processedNodes.contains(node)
                : String.format("Node %s already processed", node);

        int nodeIndex = index;
        index += 1;

        var successorIterator = successorFunction.apply(node);
        TarjanState<S> state = new TarjanState<>(node, nodeIndex, successorIterator);

        explorationStack.push(node);
        stateMap.put(node, state);
        return state;
    }

    @SuppressWarnings("ObjectEquality")
    private void run(S initial) {
        assert path.isEmpty();
        var state = create(initial);

        //noinspection LabeledStatement - Without the label this method gets ugly
        outer:
        while (true) {
            var node = state.node;
            int nodeIndex = state.nodeIndex;

            var successorIterator = state.successorIterator;
            while (successorIterator.hasNext()) {
                var successor = successorIterator.next();

                if (node.equals(successor)) {
                    if (state.lowLink == NO_LINK) {
                        state.lowLink = nodeIndex;
                    }
                    // No need to process self-loops
                    continue;
                }

                if (processedNodes.contains(successor)) {
                    continue;
                }

                var successorState = stateMap.get(successor);
                assert successorState != state; // NOPMD

                if (successorState == null) {
                    // Successor was not processed, do that now
                    path.push(state);
                    state = create(successor);
                    //noinspection ContinueStatementWithLabel
                    continue outer;
                }

                // Successor is not fully explored + we found a link to it, hence the low-link of this
                // state is less than or equal to the successors link.
                int successorIndex = successorState.nodeIndex;
                assert successorIndex != nodeIndex;

                int successorLowLink = successorState.getLowLink();
                if (successorLowLink < state.lowLink) {
                    state.lowLink = successorLowLink;
                }

                assert state.lowLink <= nodeIndex;
            }

            // Finished handling this state by identifying whether it is a root of an SCC and
            // backtracking information if not. There are three possible cases:
            // 1) No link to this state has been found at all (-> transient SCC)
            // 2) This state is its own low-link (-> root of true SCC)
            // 3) State has true low link (-> non-root element of SCC)

            int lowLink = state.lowLink;
            if (lowLink == NO_LINK) {
                // This state has no back-link at all - transient state
                assert explorationStack.peek() == node;
                assert isTransient(successorFunction, Set.of(node));

                if (this.includeTransient && !consumer.test(Set.of(node))) {
                    return;
                }

                explorationStack.pop();
                processedNodes.add(node);
            } else if (lowLink == nodeIndex) {
                // This node can't reach anything younger than itself, thus by invariant it is the root of
                // an SCC. We now build the SCC and remove all now superfluous information (to keep the used
                // data-structures as small as possible)
                assert !explorationStack.isEmpty();

                // Gather all states in this SCC by popping the stack until we find the back-link
                Set<S> scc;
                var stackNode = explorationStack.pop();
                if (stackNode.equals(node)) { // NOPMD
                    // Singleton SCC
                    scc = Set.of(node);
                    assert !isTransient(successorFunction, scc);
                } else {
                    scc = new HashSet<>();
                    scc.add(stackNode);
                    do {
                        // Pop the stack until we find our node
                        stackNode = explorationStack.pop();
                        scc.add(stackNode);
                    } while (!stackNode.equals(node)); // NOPMD
                }

                if (!consumer.test(scc)) {
                    return;
                }

                // Remove all information about the popped states - retain the index information since
                // we need to know which states have been processed.
                stateMap.keySet().removeAll(scc);
                processedNodes.addAll(scc);
            } else {
                // If this state is not a root, update the predecessor (which has to exist)
                assert !path.isEmpty() && lowLink < nodeIndex;

                var predecessorState = path.getFirst();
                // Since the current state has a "true" low-link, it is a possible low-link for the
                // predecessor, too. By invariant, it points to a non-finished state, i.e. a state in some
                // not yet found SCC.
                int predecessorLowLink = predecessorState.lowLink;

                if (lowLink < predecessorLowLink) {
                    // Also happens if predecessor's low-link is NO_LINK - we may have found a back-edge to
                    // the predecessor
                    predecessorState.lowLink = lowLink;
                }
            }

            // Backtrack on the work-stack
            if (path.isEmpty()) {
                break;
            }
            state = path.pop();
        }

        //noinspection ConstantConditions
        assert path.isEmpty();
    }

    static final class TarjanState<S> {
        final S node;
        final int nodeIndex;
        final Iterator<S> successorIterator;
        int lowLink;

        TarjanState(S node, int nodeIndex, Iterator<S> successorIterator) {
            this.node = node;
            this.nodeIndex = nodeIndex;
            this.successorIterator = successorIterator;
            lowLink = NO_LINK;
        }

        public int getLowLink() {
            // In standard Tarjan, all "NO_LINK"-states would have their own index as low-link.
            return lowLink == NO_LINK ? nodeIndex : lowLink;
        }

        @Override
        public String toString() {
            return nodeIndex + "(" + (lowLink == NO_LINK ? "X" : lowLink) + ") " + node;
        }
    }
}
